"""Test for data conversion functions."""

import time
import unittest

import networkx as nx
import numpy as np
import pytest
import scipy.sparse as sp

from qci_client.data_converter import data_to_json

@pytest.mark.offline
class TestDataToJson(unittest.TestCase):
    """Test suite for data conversion to JSON."""

    def test_file_type_assert(self):
        """Test file generation for bad file type."""
        # test file_type assertion
        with self.assertRaises(AssertionError) as context:
            data_to_json(file={
                "file_config": {"graph_partitioning_results": {"data": []}}
            })

        self.assertEqual(
           str(context.exception),
           "data conversion not supported for file type 'graph_partitioning_results', "
           "supported types are ('constraints', 'graph', 'hamiltonian', 'objective', "
           "'polynomial', 'qubo')"
        )

    def test_graph_file_body(self):
        """Test file generation for graph data."""
        graph_file = {
            "file_name": "graph.json",
            "file_config": {
                "graph": {"data": nx.Graph((((1, 2), (1, 3))))}
            },
        }
        expected = {
            "file_name": "graph.json",
            "file_config": {
                "graph": {
                    "num_edges": 2,
                    "num_nodes": 3,
                    "directed": False,
                    "multigraph": False,
                    "graph": {},
                    "nodes": [{'id': 1}, {'id': 2}, {'id': 3}],
                    "links": [{"source": 1, "target": 2}, {"source": 1, "target": 3}],
                },
            },
        }
        got = data_to_json(file=graph_file)
        self.assertDictEqual(got, expected)

        graph_file["file_config"]["graph"]["data"] = []

        with self.assertRaises(AssertionError) as context:
           data_to_json(file=graph_file)

        self.assertEqual(
           str(context.exception), "file_type 'graph' data must be a networkx.Graph"
        )

    @pytest.mark.timing
    def test_large_data_conversion(self):
        """Test file generation for large data is sufficiently fast."""
        large_qubo = np.random.normal(size=(3000, 3000))
        large_qubo = large_qubo + large_qubo.T
        start = time.perf_counter()
        data_to_json(file={"file_config": {"qubo": {"data": large_qubo}}})
        end = time.perf_counter()
        conversion_time = end - start

        self.assertTrue(
            conversion_time < 5,
            msg=f"Matrix conversion to JSON took too long: 5s <= {conversion_time}s.",
        )

    def test_type_not_graph_check(self):
        """Test file generation for mismatched data and problem types."""
        bad_hamiltonian_file = {
            "file_config": {
                "hamiltonian": {
                    "num_variables": 3,
                    "data": nx.Graph((((1, 2), (1, 3)))),
                }
            }
        }

        with self.assertRaises(AssertionError) as context:
            data_to_json(file=bad_hamiltonian_file)

        self.assertEqual(
            str(context.exception),
            "file_type 'hamiltonian' data cannot be a networkx.Graph",
        )

    def test_assert_type_not_list(self):
        """Test file generation for improperly formatted qubo data."""
        bad_qubo_file = {
            "file_config": {
                "qubo": {
                    "num_variables": 3,
                    "data": [[1, -1], [-1, 1]],
                }
            }
        }
    
        with self.assertRaises(ValueError) as context:
            data_to_json(file=bad_qubo_file)

        self.assertEqual(
            str(context.exception),
            "file_type 'qubo' only supports types numpy.ndarray and "
            "scipy.sparse.spmatrix, got <class 'list'>",
        )

    def test_hamiltonian_file_body(self):
        """Test file generation for a hamiltonian matrix."""
        expected = {
            "file_name": "hamiltonian.json",
            "file_config": {
                "hamiltonian": {
                    "num_variables": 2,
                    "data": [
                        {"i": 0, "j": 0, "val": -1.0},
                        {"i": 0, "j": 1, "val": 1.0},
                        {"i": 0, "j": 2, "val": 1.0},
                        {"i": 1, "j": 0, "val": 1.0},
                        {"i": 1, "j": 1, "val": -1.0},
                        {"i": 1, "j": 2, "val": 1.0},
                    ]
                }
            }
        }

        ham_np = np.array([[-1, 1, 1], [1, -1, 1]])

        file = {
            "file_name": "hamiltonian.json",
            "file_config": {"hamiltonian": {"data": ham_np}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

        file = {
            "file_name": "hamiltonian.json",
            "file_config": {"hamiltonian": {"data": sp.dok_matrix(ham_np)}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

    def test_qubo_file_body(self):
        """Test file generation for a qubo matrix."""
        expected = {
            "file_name": "qubo.json",
            "file_config": {
                "qubo": {
                    "num_variables": 2,
                    "data": [
                        {"i": 0, "j": 0, "val": -1.0},
                        {"i": 0, "j": 1, "val": 1.0},
                        {"i": 1, "j": 0, "val": 1.0},
                        {"i": 1, "j": 1, "val": -1.0},
                    ]
                }
            }
        }

        q_obj_np = np.array([[-1.0, 1.0], [1.0, -1.0]])

        file = {
            "file_name": "qubo.json",
            "file_config": {"qubo": {"data": q_obj_np}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

        file = {
            "file_name": "qubo.json",
            "file_config": {"qubo": {"data": sp.dok_matrix(q_obj_np)}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

    def test_objective_file_body(self):
        """Test file generation for an objective matrix."""
        expected = {
            "file_name": "objective.json",
            "file_config": {
                "objective": {
                    "num_variables": 2,
                    "data": [
                        {"i": 0, "j": 0, "val": -1.0},
                        {"i": 0, "j": 1, "val": 1.0},
                        {"i": 1, "j": 0, "val": 1.0},
                        {"i": 1, "j": 1, "val": -1.0},
                    ]
                }
            }
        }

        obj_np = np.array([[-1.0, 1.0], [1.0, -1.0]])

        file = {
            "file_name": "objective.json",
            "file_config": {"objective": {"data": obj_np}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

        file = {
            "file_name": "objective.json",
            "file_config": {"objective": {"data": sp.dok_matrix(obj_np)}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

    def test_constraints_file_body(self):
        """Test file generation for a constraints matrix."""
        expected = {
            "file_name": "constraints.json",
            "file_config": {
                "constraints": {
                    "num_constraints": 1,
                    "num_variables": 2,
                    "data": [
                        {"i": 0, "j": 0, "val": 1.0},
                        {"i": 0, "j": 1, "val": 1.0},
                        {"i": 0, "j": 2, "val": -2.0},
                    ]
                }
            }
        }

        con_np = np.array([[1.0, 1.0, -2.0]])

        file = {
            "file_name": "constraints.json",
            "file_config": {"constraints": {"data": con_np}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

        file = {
            "file_name": "constraints.json",
            "file_config": {"constraints": {"data": sp.dok_matrix(con_np)}}
        }
        self.assertDictEqual(data_to_json(file=file), expected)

    def test_polynomial_file_body(self):
        """Test file generation for a polynomial."""
        data = [
            {"idx": [0, 0, 0], "val": 1.0},
            {"idx": [0, 0, 1], "val": 2.0},
            {"idx": [0, 0, 2], "val": 3.0},
            {"idx": [0, 1, 1], "val": 4.0},
            {"idx": [0, 1, 2], "val": 5.0},
            {"idx": [0, 2, 2], "val": 6.0},
            {"idx": [1, 1, 1], "val": 7.0},
            {"idx": [1, 1, 2], "val": 8.0},
            {"idx": [1, 2, 2], "val": 9.0},
            {"idx": [2, 2, 2], "val": 10.0},
        ]

        expected = {
            "file_name": "polynomial.json",
            "file_config": {
                "polynomial": {
                    "min_degree": 0,
                    "max_degree": 3,
                    "num_variables": 2,
                    "data": data,
                }
            }
        }

        file = {
            "file_name": "polynomial.json",
            "file_config": {
                "polynomial": {
                    "min_degree": 0,
                    "max_degree": 3,
                    "num_variables": 2,
                    "data": data,
                }
            }
        }
        self.assertDictEqual(data_to_json(file=file), expected)



"""
Types for Python client for files REST API for optimization service.

Copyright 2023, Quantum Computing Incorporated
"""

from typing import Dict, List, Literal, Protocol, TypedDict, Union

# Literals should match corresponding values in enum.FileType.
Hamiltonian = Literal["hamiltonian"]
Qubo = Literal["qubo"]
Objective = Literal["objective"]
Polynomial = Literal["polynomial"]
Constraints = Literal["constraints"]
Graph = Literal["graph"]
GpResults = Literal["graph_partitioning_results"]
IhoResults = Literal["ising_hamiltonian_optimization_results"]
NqhoContinuousResults = Literal[
    "normalized_qudit_hamiltonian_optimization_continuous_results"
]
NqhoIntegerResults = Literal[
    "normalized_qudit_hamiltonian_optimization_integer_results"
]
QlcboResults = Literal["quadratic_linearly_constrained_binary_optimization_results"]
QuboResults = Literal["quadratic_unconstrained_binary_optimization_results"]


class HealthGetResponseBody(TypedDict):
    """GET health response body."""

    message: str


class VersionGetResponseBody(TypedDict):
    """GET version response body."""

    application_name: str
    version: str


class CountsGetResponseBody(TypedDict):
    """GET file counts response body."""

    count: int


class ConstraintsMetadata(TypedDict):
    """Constraints file metadata."""

    num_constraints: int
    num_variables: int


class ConstraintsMetadataConfig(TypedDict):
    """Constraints file metadata configuration."""

    constraints: ConstraintsMetadata


class GraphMetadata(TypedDict):
    """Graph file metadata."""

    directed: bool
    multigraph: bool
    graph: Dict[str, str]
    num_edges: int
    num_nodes: int


class GraphMetadataConfig(TypedDict):
    """Graph file metadata configuration."""

    graph: GraphMetadata


class HamiltonianMetadata(TypedDict):
    """Hamiltonian file metadata."""

    num_variables: int


class HamiltonianMetadataConfig(TypedDict):
    """Hamiltonian file metadata configuration."""

    hamiltonian: HamiltonianMetadata


class ObjectiveMetadata(TypedDict):
    """Objective file metadata."""

    num_variables: int


class ObjectiveMetadataConfig(TypedDict):
    """Objective file metadata configuration."""

    objective: ObjectiveMetadata


class PolynomialMetadata(TypedDict):
    """Polynomial file metadata."""

    min_degree: int
    max_degree: int
    num_variables: int


class PolynomialMetadataConfig(TypedDict):
    """Polynomial file metadata configuration."""

    polynomial: PolynomialMetadata


class QuboMetadata(TypedDict):
    """QUBO file metadata."""

    num_variables: int


class QuboMetadataConfig(TypedDict):
    """QUBO file metadata configuration."""

    qubo: QuboMetadata


class GpResultsMetadata(TypedDict):
    """Graph Partitioning (GP) results file metadata."""


class GpResultsMetadataConfig(TypedDict):
    """Graph Partitioning (GP)  results file metadata configuration."""

    graph_partitioning_results: GpResultsMetadata


class IhoResultsMetadata(TypedDict):
    """Ising Hamiltonian Optimization (IHO) results file metadata."""


class IhoResultsMetadataConfig(TypedDict):
    """Ising Hamiltonian Optimization (IHO) results file metadata configuration."""

    ising_hamiltonian_optimization_results: IhoResultsMetadata


class NqhoContinuousResultsMetadata(TypedDict):
    """
    Continuous normalized-Qudit Hamiltonian Optimization (NQHO) results file metadata.
    """


class NqhoContinuousResultsMetadataConfig(TypedDict):
    """
    Continuous Normalized-Qudit Hamiltonian Optimization (NQHO) results file metadata
    configuration.
    """

    normalized_qudit_hamiltonian_optimization_continuous_results: NqhoContinuousResultsMetadata  # pylint: disable=line-too-long


class NqhoIntegerResultsMetadata(TypedDict):
    """
    Integer normalized-Qudit Hamiltonian Optimization (NQHO) results file metadata.
    """


class NqhoIntegerResultsMetadataConfig(TypedDict):
    """
    Integer Normalized-Qudit Hamiltonian Optimization (NQHO) results file metadata
    configuration.
    """

    normalized_qudit_hamiltonian_optimization_integer_results: NqhoIntegerResultsMetadata  # pylint: disable=line-too-long


class NqhoResultsMetadata(TypedDict):
    """Normalized-Qudit Hamiltonian Optimization (NQHO) results file metadata."""


class NqhoResultsMetadataConfig(TypedDict):
    """
    Normalized-Qudit Hamiltonian Optimization (NQHO) results file metadata
    configuration.
    """

    normalized_qudit_hamiltonian_optimization_results: NqhoResultsMetadata


class QlcboResultsMetadata(TypedDict):
    """
    Quadratic Linearly Constrained Binary Optimization (QLCBO) results file metadata.
    """


class QlcboResultsMetadataConfig(TypedDict):
    """
    Quadratic Linearly Constrained Binary Optimization (QLCBO) results file metadata
    configuration.
    """

    quadratic_linearly_constrained_binary_optimization_results: QlcboResultsMetadata


class QuboResultsMetadata(TypedDict):
    """Quadratic Unconstrained Binary Optimization (QUBO) results file metadata."""


class QuboResultsMetadataConfig(TypedDict):
    """
    Quadratic Unconstrained Binary Optimization (QUBO) results file metadata
    configuration.
    """

    quadratic_unconstrained_binary_optimization_results: QuboResultsMetadata


class MetadataOptionalPostRequestBody(TypedDict, total=False):
    """Optional file metadata."""

    file_name: str


class InputMetadataPostRequestBody(MetadataOptionalPostRequestBody):
    """Request body for POSTing any input file metadata."""

    file_config: Union[
        ConstraintsMetadataConfig,
        GraphMetadataConfig,
        HamiltonianMetadataConfig,
        ObjectiveMetadataConfig,
        PolynomialMetadataConfig,
        QuboMetadataConfig,
    ]


class ResultsMetadataPostRequestBody(MetadataOptionalPostRequestBody):
    """Request body for POSTing any results file metadata."""

    organization_id: str
    user_id: str
    file_config: Union[
        GpResultsMetadataConfig,
        IhoResultsMetadataConfig,
        NqhoContinuousResultsMetadataConfig,
        NqhoIntegerResultsMetadataConfig,
        NqhoResultsMetadataConfig,
        QlcboResultsMetadataConfig,
        QuboResultsMetadataConfig,
    ]


# Request body for POSTing any file metadata.
MetadataPostRequestBody = Union[
    InputMetadataPostRequestBody,
    ResultsMetadataPostRequestBody,
]


class MetadataPostResponseBody(TypedDict):
    """File metadata POST response body."""

    file_id: str


class MatrixElement(TypedDict):
    """Matrix element dictionary."""

    i: int
    j: int
    val: float


class ConstraintsPart(TypedDict):
    """Constraints file part."""

    data: List[MatrixElement]


class ConstraintsPartConfig(TypedDict):
    """Constraints file part configuration."""

    constraints: ConstraintsPart


class NodeDict(TypedDict):
    """Dictionary elements of nodes array."""

    id: int


class LinkDictOptional(TypedDict, total=False):
    """Optional fields of dictionary elements of links array."""

    weight: float


class LinkDict(LinkDictOptional):
    """Dictionary elements of links array."""

    source: int
    target: int


class GraphPart(TypedDict):
    """Graph file part."""

    nodes: List[NodeDict]
    links: List[LinkDict]


class GraphPartConfig(TypedDict):
    """Graph file part configuration."""

    graph: GraphPart


class HamiltonianPart(TypedDict):
    """Hamiltonian file part."""

    data: List[MatrixElement]


class HamiltonianPartConfig(TypedDict):
    """Hamiltonian file part configuration."""

    hamiltonian: HamiltonianPart


class ObjectivePart(TypedDict):
    """Objective file part."""

    data: List[MatrixElement]


class ObjectivePartConfig(TypedDict):
    """Objective file part configuration."""

    objective: ObjectivePart


class PolynomialElement(TypedDict):
    """Polynomial element dictionary."""

    idx: List[int]
    val: float


class PolynomialPart(TypedDict):
    """Polynomial file part."""

    data: List[PolynomialElement]


class PolynomialPartConfig(TypedDict):
    """Polynomial file part configuration."""

    polynomial: PolynomialPart


class QuboPart(TypedDict):
    """QUBO file part."""

    data: List[MatrixElement]


class QuboPartConfig(TypedDict):
    """QUBO file part configuration."""

    qubo: QuboPart


class GpResultsPart(TypedDict):
    """GP results file part."""

    balances: List[float]
    counts: List[int]
    cut_sizes: List[float]
    energies: List[float]
    feasibilities: List[bool]
    partitions: List[List[List[int]]]
    solutions: List[List[int]]


class GpResultsPartConfig(TypedDict):
    """GP results file part configuration."""

    graph_partitioning_results: GpResultsPart


class IhoResultsPart(TypedDict):
    """IHO results file part."""

    counts: List[int]
    energies: List[float]
    solutions: List[List[int]]


class IhoResultsPartConfig(TypedDict):
    """IHO results file part configuration."""

    ising_hamiltonian_optimization_results: IhoResultsPart


class NqhoContinuousResultsPart(TypedDict):
    """Continuous NQHO results file part."""

    counts: List[int]
    energies: List[float]
    solutions: List[List[float]]


class NqhoContinuousResultsPartConfig(TypedDict):
    """Continuous NQHO results file part configuration."""

    normalized_qudit_hamiltonian_optimization_continuous_results: NqhoContinuousResultsPart  # pylint: disable=line-too-long


class NqhoIntegerResultsPart(TypedDict):
    """Integer NQHO results file part."""

    counts: List[int]
    energies: List[float]
    solutions: List[List[int]]


class NqhoIntegerResultsPartConfig(TypedDict):
    """Integer NQHO results file part configuration."""

    normalized_qudit_hamiltonian_optimization_integer_results: NqhoIntegerResultsPart


class NqhoResultsPartWithoutDistilled(TypedDict):
    """NQHO results file part without distilled solution results."""

    counts: List[int]
    energies: List[float]
    solutions: List[List[int]]


class NqhoResultsPartWithDistilled(NqhoResultsPartWithoutDistilled):
    """NQHO results file part with distilled solution results."""

    distilled_energies: List[float]
    distilled_solutions: List[List[int]]


class NqhoResultsPartConfig(TypedDict):
    """NQHO results file part configuration."""

    normalized_qudit_hamiltonian_optimization_results: Union[
        NqhoResultsPartWithoutDistilled, NqhoResultsPartWithDistilled,
    ]


class QlcboResultsPart(TypedDict):
    """QLCBO results file part."""

    counts: List[int]
    energies: List[float]
    feasibilities: List[bool]
    objective_values: List[float]
    solutions: List[List[int]]


class QlcboResultsPartConfig(TypedDict):
    """QLCBO results file part configuration."""

    quadratic_linearly_constrained_binary_optimization_results: QlcboResultsPart


class QuboResultsPart(TypedDict):
    """QUBO results file part."""

    counts: List[int]
    energies: List[float]
    solutions: List[List[int]]


class QuboResultsPartConfig(TypedDict):
    """QUBO results file part configuration."""

    quadratic_unconstrained_binary_optimization_results: QuboResultsPart


class PartPatchRequestBody(TypedDict):
    """The request body for PATCHing all input and results file parts."""

    file_config: Union[
        ConstraintsPartConfig,
        GraphPartConfig,
        HamiltonianPartConfig,
        ObjectivePartConfig,
        PolynomialPartConfig,
        QuboPartConfig,
        GpResultsPartConfig,
        IhoResultsPartConfig,
        NqhoContinuousResultsPartConfig,
        NqhoIntegerResultsPartConfig,
        NqhoResultsPartConfig,
        QlcboResultsPartConfig,
        QuboResultsPartConfig,
    ]


class PartPatchResponseBody(TypedDict):
    """File part PATCH response body."""

    file_id: str


class MetadataOptionalGetResponseBody(TypedDict, total=False):
    """Optional fields for file metadata GET responses."""

    organization_id: str
    user_id: str
    file_name: str


class MetadataGetResponseBody(MetadataOptionalGetResponseBody):
    """GET file metatada response body."""

    file_id: str
    upload_date_rfc3339: str
    last_accessed_rfc3339: str
    num_parts: int
    num_bytes: int
    file_config: Union[
        ConstraintsMetadataConfig,
        GraphMetadataConfig,
        HamiltonianMetadataConfig,
        ObjectiveMetadataConfig,
        QuboMetadataConfig,
        GpResultsMetadataConfig,
        IhoResultsMetadataConfig,
        NqhoContinuousResultsMetadataConfig,
        NqhoIntegerResultsMetadataConfig,
        NqhoResultsMetadataConfig,
        QlcboResultsMetadataConfig,
        QuboResultsMetadataConfig,
    ]


class GetResponseBody(TypedDict):
    """Fields for GET file metadata responses."""

    files: List[MetadataGetResponseBody]


class PartGetResponseBody(TypedDict):
    """GET file part response body."""

    file_id: str
    part_number: int
    upload_date_rfc3339: str
    file_config: Union[
        ConstraintsPartConfig,
        GraphPartConfig,
        HamiltonianPartConfig,
        ObjectivePartConfig,
        QuboPartConfig,
        GpResultsPartConfig,
        IhoResultsPartConfig,
        NqhoContinuousResultsPartConfig,
        NqhoIntegerResultsPartConfig,
        NqhoResultsPartConfig,
        QlcboResultsPartConfig,
        QuboResultsPartConfig,
    ]


class DeleteResponseBody(TypedDict):
    """File DELETE response body."""

    num_deleted: int


class SupportsAccessTokens(Protocol):  # pylint: disable=too-few-public-methods
    """Interface for clients that are injected to retrieve access tokens."""

    @property
    def access_token(self) -> str:
        """Return a valid access token for the API."""
        ...  # pylint: disable=unnecessary-ellipsis



"""Test suite for jobs run on a remote backend."""

from copy import deepcopy
from typing import Optional
import unittest

import numpy as np
import scipy.sparse as sp

from qci_client import JobStatus
import qci_client.base
from tests.test_qci_client import TestQciClientFiles


class TestQciClientRemoteJobs(TestQciClientFiles):
    """Collection of tests for remote jobs."""

    def run_end_end(
        self, job_type: str, job_body: dict, result_keys: Optional[list] = None
    ):
        """
        Utility function for testing end to end pipeline.
        :param job_type: one of _supported_job_types
        :param job_body: a validate job request
        Testing each in series:
            - submit_job
            - get_job_status
            - get_job_response
            - get_file
        """

        result_keys = self.result_keys if result_keys is None else result_keys

        job_id = self.q1.submit_job(job_body=job_body, job_type=job_type)

        self.assertIn("job_id", job_id)
        self.assertIsInstance(job_id["job_id"], str)

        status = {"status": ""}

        while status["status"] not in self.final_status:
            status = self.q1.get_job_status(job_id=job_id["job_id"])
            self.assertIn(status["status"], self.all_statuses)

        self.assertEqual(JobStatus.COMPLETED, status["status"])
        response = self.q1.get_job_response(job_id=job_id["job_id"], job_type=job_type)

        # test job info has appropriate keys
        # self.assertEqual(self.job_info, set(response.keys()))  # FIXME

        result = self.q1.download_file(file_id=response["job_result"]["file_id"])
        result = list(result)
        # self.assertTrue(all(i in result[0] for i in result_keys))  # FIXME


    def process_job_check(self, job_type, job_body):
        """Utility function for checking job types."""
        process_key = ["job_info", "results"]
        job_output = self.q1.process_job(
            job_type=job_type, job_body=job_body, wait=True
        )
        self.assertTrue(all(key in process_key for key in list(job_output.keys())))

    def test_large_job(self):
        """Test large sample-qubo job."""
        num_variables = 1000
        large_qubo = sp.random(
               num_variables, num_variables, density=0.5, format='dok', dtype=np.float32
        )
        large_qubo = (large_qubo + large_qubo.T) / 2
        large_qubo_dict = {
            "file_name": "test_large_qubo",
            "file_config": {
                "qubo": {
                    "data": large_qubo
                }
            }
        }
        large_file_id = self.q1.upload_file(file=large_qubo_dict)["file_id"]
        print("LARGE FILE ID", large_file_id)
        large_job_body = {
            "job_submission": {
                "job_name": "large_qubo_test_job",
                "problem_config": {
                    "quadratic_unconstrained_binary_optimization": {
                        "qubo_file_id": large_file_id
                    }
                },
                "device_config": {
                    "dirac-1": {}
                }
            }
        }

        self.process_job_check(job_type="sample-qubo", job_body=large_job_body)

    def test_process_qubo(self):
        """Test that sample-qubo job process can be checked."""
        self.process_job_check(job_type="sample-qubo", job_body=self.qubo_job_body)

    def test_process_constraint(self):
        """Test that sample-constraint job process can be checked."""
        self.process_job_check(
            job_type="sample-constraint", job_body=self.constraint_job_body
        )

    def test_process_hamiltonian_ising_dirac1(self):
        """Test that sample-hamiltonian-ising process job on dirac-1 can be checked."""
        self.process_job_check(
            job_type="sample-hamiltonian-ising",
            job_body=self.hamiltonian_job_body_ising_dirac1,
        )

    def test_process_hamiltonian_dirac2(self):
        """Test that integer sample-hamiltonian process job on dirac-2 can be checked."""
        self.process_job_check(
            job_type="sample-hamiltonian",
            job_body=self.hamiltonian_job_body_integer_dirac2,
        )

    def test_process_hamiltonian_dirac3(self):
        """
        Test that undistilled sample-hamiltonian process on dirac-3 job can be checked.
        """
        self.process_job_check(
            job_type="sample-hamiltonian",
            job_body=self.hamiltonian_job_body_undistilled_dirac3,
        )

    def test_graph_partitioning(self):
        """Test graph-partitioning job."""
        self.run_end_end(job_type="graph-partitioning", job_body=self.graph_job_body)

    def test_sample_qubo(self):
        """Test sample-qubo job."""
        self.run_end_end(job_type="sample-qubo", job_body=self.qubo_job_body)

    def test_sample_constraint(self):
        """Test sample-constraint job."""
        self.run_end_end(
            job_type="sample-constraint",
            job_body=self.constraint_job_body,
        )

    def test_sample_hamiltonian_ising_polynomial_dirac1(self):
        """Test sample-hamiltonian-ising job using polynomial on dirac-1."""
        self.run_end_end(
            job_type="sample-hamiltonian-ising",
            job_body=self.polynomial_job_body_ising_dirac1
        )

    def test_sample_hamiltonian_ising_hamiltonian_dirac1(self):
        """Test sample-hamiltonian-ising job using hamiltonian matrix on dirac-1."""
        self.run_end_end(
            job_type="sample-hamiltonian-ising",
            job_body=self.hamiltonian_job_body_ising_dirac1
        )

    def test_sample_hamiltonian_polynomial_continuous_dirac2(self):
        """Test sample-hamiltonian continuous job using polynomial on dirac-2."""
        self.run_end_end(
            job_type="sample-hamiltonian",
            job_body=self.polynomial_job_body_hamiltonian_continuous_dirac2
        )

    def test_sample_hamiltonian_polynomial_integer_dirac2(self):
        """Test sample-hamiltonian integer job using polynomial on dirac-2."""
        self.run_end_end(
            job_type="sample-hamiltonian",
            job_body=self.polynomial_job_body_hamiltonian_integer_dirac2
        )

    def test_sample_hamiltonian_continuous_dirac2(self):
        """Test continuous sample-hamiltonian job on dirac-2."""
        self.run_end_end(
            job_type="sample-hamiltonian",job_body=self.hamiltonian_job_body_continous_dirac2
        )

    def test_sample_hamiltonian_integer_dirac2(self):
        """Test integer sample-hamiltonian job on dirac-2."""
        self.run_end_end(
            job_type="sample-hamiltonian",
            job_body=self.hamiltonian_job_body_integer_dirac2
        )

    def test_sample_hamiltonian_polynomial_distilled_continuous_dirac3(self):
        """
        Test distilled-to-continuous sample-hamiltonian job using polynomial on dirac-3.
        """
        self.run_end_end(
            job_type="sample-hamiltonian",
            job_body=self.polynomial_min_degree_2_max_degree_4_job_body_distilled_continuous_dirac3,
        )

    def test_sample_hamiltonian_undistilled_dirac3(self):
        """Test undistilled sample-hamiltonian job on dirac-3."""
        self.run_end_end(
            job_type="sample-hamiltonian",
            job_body=self.hamiltonian_job_body_undistilled_dirac3,
        )

    def test_sample_hamiltonian_distilled_continuous_dirac3(self):
        """Test distilled-to-continuous sample-hamiltonian job on dirac-3."""
        self.run_end_end(
            job_type="sample-hamiltonian",
            job_body=self.hamiltonian_job_body_distilled_continuous_dirac3,
        )

    def test_sample_hamiltonian_distilled_integer_dirac3(self):
        """Test distilled-to-integer sample-hamiltonian job on dirac-3."""
        self.run_end_end(
            job_type="sample-hamiltonian",
            job_body=self.hamiltonian_job_body_distilled_integer_dirac3,
        )


@qci_client.base.BaseApi.refresh_token
def refresh_token_wrapped_function(*_, **__) -> None:
    """Mock function wrapped with token checking/fetching."""


class TestBaseApi(unittest.TestCase):
    """BaseApi-related test suite."""

    def test_is_bearer_token_expired_uninit(self):
        """Test is_bearer_token_expired when not retrieved during intialization."""
        # This should initialize with unset bearer info that is considered to have expired token.
        base_api = qci_client.base.BaseApi(set_bearer_token_on_init=False)
        self.assertTrue(base_api.is_bearer_token_expired())

        # Calling this function should set bearer info expiration.
        refresh_token_wrapped_function(base_api)
        self.assertFalse(base_api.is_bearer_token_expired())
        original_bearer_info = deepcopy(base_api._bearer_info)

        # Calling this function again should not reset set bearer info expiration.
        refresh_token_wrapped_function(base_api)
        self.assertFalse(base_api.is_bearer_token_expired())
        self.assertDictEqual(base_api._bearer_info, original_bearer_info)

    def test_is_bearer_token_expired_init(self):
        """Test is_bearer_token_expired when retrieved during intialization."""
        # This should initialize with set bearer info with unexpired token.
        base_api = qci_client.base.BaseApi()
        self.assertFalse(base_api.is_bearer_token_expired())
        original_bearer_info = deepcopy(base_api._bearer_info)

        # Calling this function again should not reset set bearer info expiration.
        refresh_token_wrapped_function(base_api)
        self.assertFalse(base_api.is_bearer_token_expired())
        self.assertDictEqual(base_api._bearer_info, original_bearer_info)



"""Base class for API clients."""

from dataclasses import dataclass, field
from datetime import datetime
import functools
import os
from posixpath import join
from typing import Callable, Optional

import requests
from requests.adapters import HTTPAdapter, Retry

# We are uploading files we want to retry when we receive certain error codes
RETRY_TOTAL = 7
BACKOFF_FACTOR = 2
STATUS_FORCELIST = [502, 503, 504]


@dataclass
class BaseApi:  # pylint: disable=too-many-instance-attributes
    """
    Base class for clients to QCi APIs, especially the authorization layer.

    :param url: url basepath to API endpoint, including scheme, if None, then falls back
        to QCI_API_URL environment variable
    :param api_token: refresh token for authenticating to API, if None, then falls back
        to QCI_TOKEN environment variable
    :param access_tokens: url path fragment to specify access-tokens API endpoint
    :param set_bearer_token_on_init: flag to turn on/off access token retrieval on
        object initialization
    :param timeout: number of seconds before timing out requests, None waits indefinitely
    :param debug: flag to turn on/off debug prints
    """

    url: Optional[str] = None
    # Hide sensistive info to prevent accidental logging when printing client objects.
    api_token: Optional[str] = field(default=None, repr=False)
    access_tokens: str = "auth/v1/access-tokens"
    set_bearer_token_on_init: bool = True
    # Request timeout in seconds for connection & read. None for infinite timeout.
    timeout: Optional[float] = None
    debug: bool = False

    def __post_init__(self):
        self.url = self.url if self.url else os.getenv("QCI_API_URL")

        if self.url is None:
            raise AssertionError(
                "QCI_API_URL environment variable is empty. Specify url or add the "
                "necessary environment variable"
            )

        # removing trailling / so can add paths simply
        self.url.rstrip("/")

        self.api_token = self.api_token if self.api_token else os.getenv("QCI_TOKEN")

        if self.api_token is None:
            raise AssertionError(
                "QCI_TOKEN environment variable is empty. Specify api_token or add the "
                "necessary environment variable"
            )

        self.session = requests.Session()
        retries = Retry(
            total=RETRY_TOTAL,
            backoff_factor=BACKOFF_FACTOR,
            status_forcelist=STATUS_FORCELIST,
        )
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

        self._bearer_info = {}

        if self.set_bearer_token_on_init:
            self.set_bearer_token()

    @property
    def auth_url(self) -> str:
        """Return the URL used for authorization."""
        return join(self.url, self.access_tokens)

    @property
    def headers_without_token(self):
        """Headers without cached bearer token."""
        headers = {
            "Content-Type": "application/json",
            "Connection": "close",
        }

        if self.timeout is not None:
            # Provide client's request timeout to server (latency provides some buffer).
            headers.update({"X-Request-Timeout-Nano": str(int(10**9 * self.timeout))})

        return headers

    @property
    def headers(self):
        """Headers with cached bearer token."""
        headers = self.headers_without_token
        headers["Authorization"] = f"Bearer {self._bearer_info.get('access_token', '')}"

        return headers

    @property
    def headers_without_connection_close(self):
        """Headers with cached bearer token, but without connection closing."""
        headers = self.headers
        headers.pop("Connection", None)

        return headers

    def _check_response_error(self, response: requests.Response) -> None:
        """
        Single place to update error check and message for API calls
        :param response: a response from any API call using the requests package
        """
        if response.status_code >= 300:
            raise AssertionError(
                f"Error: {response.text}. Received error code {response.status_code}"
            )

    def get_bearer_token(self) -> requests.Response:
        """Request new bearer token. (Not cached here, see set_bearer_token.)"""
        payload = {"refresh_token": self.api_token}

        response = self.session.request(
            "POST",
            self.auth_url,
            json=payload,
            headers=self.headers_without_token,
            timeout=self.timeout,
        )

        self._check_response_error(response)

        return response

    def set_bearer_token(self) -> None:
        """Set bearer token from request."""
        resp = self.get_bearer_token()
        self._bearer_info = resp.json()

    def is_bearer_token_expired(self) -> bool:
        """Is current time > 'expires' time."""
        if "expires_at_rfc3339" not in self._bearer_info:
            return True

        expiration = datetime.strptime(
            self._bearer_info["expires_at_rfc3339"], "%Y-%m-%dT%H:%M:%SZ"
        )
        seconds_to_expiration = (expiration - datetime.utcnow()).total_seconds()

        # adding 10 second buffer for expiration
        return seconds_to_expiration < 10

    @staticmethod
    def refresh_token(func) -> Callable:
        """Return a wrapper function that can check an auth token."""

        @functools.wraps(func)
        def check_token(api, *args, **kwargs):
            # Because the decorated function is receiving 'self', we need to pass this
            # additional argument along in the 'api' arg.
            is_expired = api.is_bearer_token_expired()
            # expired, reset the token
            if is_expired:
                api.set_bearer_token()
                return func(api, *args, **kwargs)
            # still have time on the token, so just pass the wrapped func through
            return func(api, *args, **kwargs)

        return check_token



"""
Enumerations for Python client for files REST API for optimization service.

Copyright 2023, Quantum Computing Incorporated
"""

from enum import Enum


class FileType(Enum):
    """Enumeration of all file types."""

    # Enums should match corresponding literals in types module.
    # Job input files.
    CONSTRAINTS = "constraints"
    GRAPH = "graph"
    HAMILTONIAN = "hamiltonian"
    OBJECTIVE = "objective"
    POLYNOMIAL = "polynomial"
    QUBO = "qubo"
    # Job results files.
    GP_RESULTS = "graph_partitioning_results"
    IHO_RESULTS = "ising_hamiltonian_optimization_results"
    NQHO_CONTINUOUS_RESULTS = (
        "normalized_qudit_hamiltonian_optimization_continuous_results"
    )
    NQHO_INTEGER_RESULTS = "normalized_qudit_hamiltonian_optimization_integer_results"
    QLCBO_RESULTS = "quadratic_linearly_constrained_binary_optimization_results"
    QUBO_RESULTS = "quadratic_unconstrained_binary_optimization_results"


FILE_TYPES = frozenset(type.value for type in FileType)
JOB_INPUTS_FILE_TYPES = frozenset(
    type for type in FileType if "results" not in type.value
)
JOB_INPUTS_MATRIX_FILE_TYPES = JOB_INPUTS_FILE_TYPES - frozenset([FileType.GRAPH, FileType.POLYNOMIAL])
JOB_INPUTS_NON_GRAPH_FILE_TYPES = JOB_INPUTS_FILE_TYPES - frozenset([FileType.GRAPH])
JOB_RESULTS_FILE_TYPES = frozenset(type for type in FileType if "results" in type.value)


def get_file_type(file: dict) -> FileType:
    """Get file type from a file."""
    file_config_keys = list(file["file_config"].keys())

    if len(file_config_keys) != 1:
        raise ValueError(
            "improper number of files specified in file_config (should be exactly one)"
        )

    return FileType(file_config_keys[0])



"""
Missing tests:
 - failures should be accounted for under the BaseApi
 - Each job should also be covered for process_job only 3/5 currently are.
"""

from contextlib import redirect_stdout
from datetime import datetime
import io
from math import ceil
from typing import Optional
import unittest
import unittest.mock

import networkx as nx
import numpy as np
import pytest
import requests
import scipy.sparse as sp

from qci_client import JobStatus, QciClient, compute_results_step_len


@pytest.mark.offline
class TestJobStatus(unittest.TestCase):
    """JobStatus-related test suite."""

    def test_jobstatus(self):
        """Test all job statuses."""
        job_status = JobStatus()
        self.assertEqual(job_status.QUEUED, "QUEUED")
        self.assertEqual(job_status.SUBMITTED, "SUBMITTED")
        self.assertEqual(job_status.RUNNING, "RUNNING")
        self.assertEqual(job_status.COMPLETED, "COMPLETED")
        self.assertEqual(job_status.ERRORED, "ERRORED")
        self.assertEqual(job_status.CANCELLED, "CANCELLED")


class TestQciClientFiles(unittest.TestCase):
    """Files-API-related test suite."""

    @classmethod
    def setUpClass(cls):
        cls.q1 = QciClient()
        cls.graph_dict_input = {
            "file_name": "graph.json",
            "file_config": {
                "graph": {
                    "data": nx.Graph(((1, 2), (1, 3)))
                }
            }
        }

        cls.qubo_dict_input = {
            "file_name": "qubo.json",
            "file_config": {
                "qubo": {
                    "data": sp.csr_matrix(
                        [
                            [-1.0, 1.0],
                            [1.0, -1.0]
                        ]
                    )
                }
            }
        }

        cls.objective_dict_input = {
            "file_name": "objective.json",
            "file_config": {
                "objective": {
                    "data": sp.csc_matrix(
                        [
                            [-1.0, 1.0],
                            [1.0, -1.0]
                        ]
                    )
                }
            }
        }

        cls.constraint_dict_input = {
            "file_name": "constraints.json",
            "file_config": {
                "constraints": {
                    "data": sp.coo_matrix(
                        [
                            [-1.0, 1.0, 1.0],
                        ]
                    )
                }
            }
        }

        cls.hamiltonian_dict_input = {
            "file_name": "hamiltonian.json",
            "file_config": {
                "hamiltonian": {
                    "data": np.array(
                        [
                            [1.0, 1.0, 0.0],
                            [2.0, 0.0, 1.0]
                        ]
                    )
                }
            }
        }

        cls.polynomial_dict_hamiltonian_input = {
            "file_name": "polynomial-hamiltonian.json",
            "file_config": {
                "polynomial": {
                    "min_degree": 1,
                    "max_degree": 2,
                    "num_variables": 2,
                    "data": [
                        {
                            "idx": [0, 1],
                            "val": 1.0
                        },
                        {
                            "idx": [1, 1],
                            "val": -2.0
                        },
                        {
                            "idx": [1, 2],
                            "val": 1.0
                        },
                        {
                            "idx": [2, 2],
                            "val": -1.0
                        }

                    ]
                }
            }
        }

        cls.polynomial_dict_input = {
            "file_name": "polynomial.json",
            "file_config": {
                "polynomial": {
                    "min_degree": 2,
                    "max_degree": 4,
                    "num_variables": 2,
                    "data": [
                        {
                            "idx": [0, 0, 1, 1],
                            "val": 1.0
                        },
                        {
                            "idx": [0, 1, 1, 1],
                            "val": -2.0
                        },
                        {
                            "idx": [1, 1, 1, 1],
                            "val": 1.0
                        }
                    ]
                }
            }
        }

        cls.graph_file_id = cls.q1.upload_file(file=cls.graph_dict_input)["file_id"]
        cls.qubo_file_id = cls.q1.upload_file(file=cls.qubo_dict_input)["file_id"]
        cls.objective_file_id = cls.q1.upload_file(file=cls.objective_dict_input)[
            "file_id"
        ]
        cls.constraints_file_id = cls.q1.upload_file(file=cls.constraint_dict_input)[
            "file_id"
        ]
        cls.hamiltonian_file_id = cls.q1.upload_file(file=cls.hamiltonian_dict_input)[
            "file_id"
        ]
        cls.polynomial_hamiltonian_file_id = cls.q1.upload_file(
            file=cls.polynomial_dict_hamiltonian_input
        )["file_id"]
        cls.polynomial_min_degree_2_max_degree_4_file_id = cls.q1.upload_file(
            file=cls.polynomial_dict_input)["file_id"]

        cls.all_statuses = [
            JobStatus.QUEUED,
            JobStatus.SUBMITTED,
            JobStatus.RUNNING,
            JobStatus.COMPLETED,
            JobStatus.ERRORED,
            JobStatus.CANCELLED,
        ]

        cls.final_status = [JobStatus.COMPLETED, JobStatus.ERRORED, JobStatus.CANCELLED]

        cls.job_info = set(("job_id", "details", "job_results", "job_submission",))
        cls.result_keys = ["samples", "energies", "file_name", "file_type"]

        cls.graph_job_body = {
            "job_submission": {
                "job_name": "job_0",
                "problem_config": {
                    "graph_partitioning": {
                        "graph_file_id": cls.graph_file_id
                    },
                },
                "device_config": {
                    "dirac-1": {}
                }
            }
        }

        cls.qubo_job_body = {
            "job_submission": {
                "job_name": "job_0",
                "problem_config": {
                    "quadratic_unconstrained_binary_optimization": {
                        "qubo_file_id": cls.qubo_file_id
                    },
                },
                "device_config": {
                    "dirac-1": {}
                }
            }
        }

        cls.constraint_job_body = {
            "job_submission": {
                "job_name": "job_0",
                "problem_config": {
                    "quadratic_linearly_constrained_binary_optimization": {
                        "objective_file_id": cls.objective_file_id,
                        "constraints_file_id": cls.constraints_file_id
                    }
                },
                "device_config": {
                    "dirac-1": {}
                }
            }
        }

        cls.hamiltonian_job_body_ising_dirac1 = {
            "job_submission": {
                "job_name": "dirac-1",
                "problem_config": {
                    "ising_hamiltonian_optimization": {
                        "hamiltonian_file_id": cls.hamiltonian_file_id
                    },
                },
                "device_config": {
                    "dirac-1": {
                        "num_samples": 2,
                    }
                }
            }
        }

        cls.polynomial_job_body_ising_dirac1 = {
            "job_submission": {
                "job_name": "dirac-1",
                "problem_config": {
                    "ising_hamiltonian_optimization": {
                        "polynomial_file_id": cls.polynomial_hamiltonian_file_id
                    },
                },
                "device_config": {
                    "dirac-1": {
                    }
                }
            }
        }

        cls.hamiltonian_job_body_continous_dirac2 = {
            "job_submission": {
                "job_name": "dirac-2",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization_continuous": {
                        "hamiltonian_file_id": cls.hamiltonian_file_id
                    }
                },
                "device_config": {
                    "dirac-2": {}
                }
            }
        }

        cls.hamiltonian_job_body_integer_dirac2 = {
            "job_submission": {
                "job_name": "dirac-2",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization_integer": {
                        "hamiltonian_file_id": cls.hamiltonian_file_id
                    },
                },
                "device_config": {
                    "dirac-2": {}
                }
            }
        }

        cls.polynomial_job_body_hamiltonian_continuous_dirac2 = {
            "job_submission": {
                "job_name": "dirac-2",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization_continuous": {
                        "polynomial_file_id": cls.polynomial_hamiltonian_file_id
                    },
                },
                "device_config": {
                    "dirac-2": {}
                }
            }
        }

        cls.polynomial_job_body_hamiltonian_integer_dirac2 = {
            "job_submission": {
                "job_name": "dirac-2",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization_integer": {
                        "polynomial_file_id": cls.polynomial_hamiltonian_file_id
                    },
                },
                "device_config": {
                    "dirac-2": {
                        "num_samples": 2,
                    }
                }
            }
        }

        cls.hamiltonian_job_body_undistilled_dirac3 = {
            "job_submission": {
                "job_name": "dirac-3",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization": {
                        "hamiltonian_file_id": cls.hamiltonian_file_id
                    }
                },
                "device_config": {
                    "dirac-3": {
                        "relaxation_parameter": 1,
                        "sum_constraint": 2.4,
                    }
                }
            }
        }

        cls.hamiltonian_job_body_distilled_continuous_dirac3 = {
            "job_submission": {
                "job_name": "dirac-3",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization": {
                        "hamiltonian_file_id": cls.hamiltonian_file_id
                    }
                },
                "device_config": {
                    "dirac-3": {
                        "relaxation_parameter": 1,
                        "sum_constraint": 2.4,
                        "solution_precision": 0.1,
                    }
                }
            }
        }

        cls.hamiltonian_job_body_distilled_integer_dirac3 = {
            "job_submission": {
                "job_name": "dirac-3",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization": {
                        "hamiltonian_file_id": cls.hamiltonian_file_id
                    }
                },
                "device_config": {
                    "dirac-3": {
                        "relaxation_parameter": 1,
                        "solution_precision": 1,
                    }
                }
            }
        }

        cls.polynomial_min_degree_2_max_degree_4_job_body_distilled_continuous_dirac3 = {
            "job_submission": {
                "job_name": "dirac-3",
                "problem_config": {
                    "normalized_qudit_hamiltonian_optimization": {
                        "polynomial_file_id": cls.polynomial_min_degree_2_max_degree_4_file_id
                    },
                },
                "device_config": {
                    "dirac-3": {
                        "relaxation_parameter": 2,
                        "sum_constraint": 2.4,
                        "solution_precision": 0.1,
                    }
                }
            }
        }

    def test_upload_file_error(self):
        """Test uploading improperly formatted file."""
        with self.assertRaises(KeyError):
            error_input = {
                "file_type": "graph",
                "file_name": "qubo.json",
                "data": [
                    {"i": 0, "j": 0, "val": -1.0},
                    {"i": 0, "j": 1, "val": 1.0},
                    {"i": 1, "j": 0, "val": 1.0},
                    {"i": 1, "j": 1, "val": -1.0},
                ],
                "num_variables": 2,
            }
            self.q1.upload_file(file=error_input)

    def test_upload_file(self):
        """Test uploading of various file types."""
        graph_upload = self.q1.upload_file(file=self.graph_dict_input)
        self.assertIn("file_id", graph_upload)
        self.assertIsInstance(graph_upload["file_id"], str)

        qubo_upload = self.q1.upload_file(file=self.qubo_dict_input)
        self.assertIn("file_id", qubo_upload)
        self.assertIsInstance(qubo_upload["file_id"], str)

        objective_upload = self.q1.upload_file(file=self.objective_dict_input)
        self.assertIn("file_id", objective_upload)
        self.assertIsInstance(objective_upload["file_id"], str)

        constraint_upload = self.q1.upload_file(file=self.constraint_dict_input)
        self.assertIn("file_id", constraint_upload)
        self.assertIsInstance(constraint_upload["file_id"], str)

        hamiltonian_upload = self.q1.upload_file(file=self.hamiltonian_dict_input)
        self.assertIn("file_id", hamiltonian_upload)
        self.assertIsInstance(hamiltonian_upload["file_id"], str)

    def test_validate_job_type(self):
        """Test validation of job type against whitelist."""
        test_job_type = "fake_job_type"
        expected_err_str = f"Provided job_type '{test_job_type}' is not one of {self.q1._supported_job_types}"  # pylint: disable=protected-access

        with self.assertRaises(AssertionError) as context:
            self.q1.validate_job_type(job_type=test_job_type)

        # TODO: consider asserting on exception type instead of specific text of assert msg
        # because aserting on the specific text of the assert msg makes this test fairly fragile
        self.assertEqual(
            str(context.exception),
            expected_err_str
        )

        try:
            self.q1.validate_job_type(job_type="sample-qubo")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(
                f"validate_job_type failed with exception '{exc}' with supported job_type"
            )

    def test_build_job_body(self):
        """Test building of various jobs' request bodies."""
        with self.assertRaises(ValueError) as context:
            self.q1.build_job_body(
                job_type="sample-qubo",
                job_params={},
                qubo_file_id="qubo_file_id",
            )

        self.assertEqual(
            str(context.exception),
            "Must define sampler_type in job_params (dirac-1, dirac-2, or dirac-3)."
        )

        constraint_body = self.q1.build_job_body(
            job_type="sample-constraint",
            job_params={
                "sampler_type": "dirac-1",
                "alpha": 5.0,
            },
            objective_file_id="obj_fid",
            constraints_file_id="cons_fid",
            job_name="foobar_name",
            job_tags=["foobar_tag1", "foobar_tag2"],
        )

        self.assertDictEqual(
            constraint_body,
            {
                "job_submission": {
                    "job_name": "foobar_name",
                    "job_tags": ["foobar_tag1", "foobar_tag2"],
                    "problem_config": {
                        "quadratic_linearly_constrained_binary_optimization": {
                            "objective_file_id": "obj_fid",
                            "constraints_file_id": "cons_fid",
                            "alpha": 5.0
                        }
                    },
                    "device_config": {
                        "dirac-1": {}
                    }
                }
            }
        )

        qubo_body = self.q1.build_job_body(
            job_type="sample-qubo",
            job_params={
                "nsamples": 24,
                "sampler_type": "dirac-1",
            },
            qubo_file_id="qubo_fid",
            job_name="foobar_name",
        )

        self.assertDictEqual(
            qubo_body,
            {
                "job_submission": {
                    "job_name": "foobar_name",
                    "problem_config": {
                        "quadratic_unconstrained_binary_optimization": {
                            "qubo_file_id": "qubo_fid",
                        }
                    },
                    "device_config": {
                        "dirac-1": {
                            "num_samples": 24
                        }
                    }
                }
            }
        )

        hamiltonian_body = self.q1.build_job_body(
            job_type="sample-hamiltonian",
            job_params={
                "n_samples": 24,
                "sampler_type": "dirac-3",
                "sum_constraint": 200,
                "solution_precision": 1,
            },
            hamiltonian_file_id="hamiltonian_fid",
            job_tags=["foobar_tag1", "foobar_tag2"],
        )

        self.assertDictEqual(
            hamiltonian_body,
            {
                "job_submission": {
                    "job_tags": ["foobar_tag1", "foobar_tag2"],
                    "problem_config": {
                        "normalized_qudit_hamiltonian_optimization": {
                            "hamiltonian_file_id": "hamiltonian_fid",
                        }
                    },
                    "device_config": {
                        "dirac-3": {
                            "num_samples": 24,
                            "sum_constraint": 200,
                            "solution_precision": 1,
                        }
                    }
                }
            }
        )

        hamiltonian_body = self.q1.build_job_body(
            job_type="sample-hamiltonian",
            job_params={
                "sampler_type": "dirac-3",
                "sum_constraint": 200,
                "solution_precision": 1,
            },
            polynomial_file_id="polynomial_fid"
        )

        self.assertDictEqual(
            hamiltonian_body,
            {
                "job_submission": {
                    "problem_config": {
                        "normalized_qudit_hamiltonian_optimization": {
                            "polynomial_file_id": "polynomial_fid",
                        }
                    },
                    "device_config": {
                        "dirac-3": {
                            "sum_constraint": 200,
                            "solution_precision": 1,
                        }
                    }
                }
            }
        )

        hamiltonian_body = self.q1.build_job_body(
            job_type="sample-hamiltonian",
            job_params={
                "sampler_type": "dirac-3",
                "sum_constraint": 2.4
            },
            polynomial_file_id="polynomial_fid"
        )

        self.assertDictEqual(
            hamiltonian_body,
            {
                "job_submission": {
                    "problem_config": {
                        "normalized_qudit_hamiltonian_optimization": {
                            "polynomial_file_id": "polynomial_fid",
                        }
                    },
                    "device_config": {
                        "dirac-3": {
                            "sum_constraint": 2.4,
                        }
                    }
                }
            }
        )

        hamiltonian_body = self.q1.build_job_body(
            job_type="sample-hamiltonian",
            job_params={
                "sampler_type": "dirac-3",
                "sum_constraint": 2.4,
                "solution_precision": 0.1,
            },
            hamiltonian_file_id="hamiltonian_fid"
        )

        self.assertDictEqual(
            hamiltonian_body,
            {
                "job_submission": {
                    "problem_config": {
                        "normalized_qudit_hamiltonian_optimization": {
                            "hamiltonian_file_id": "hamiltonian_fid",
                        }
                    },
                    "device_config": {
                        "dirac-3": {
                            "sum_constraint": 2.4,
                            "solution_precision": 0.1,
                        }
                    }
                }
            }
        )

        graph_body = self.q1.build_job_body(
            job_type="graph-partitioning",
            job_params={
                "sampler_type": "dirac-1",
                "num_samples": 42,
            },
            graph_file_id="graph_fid"
        )

        self.assertDictEqual(
            graph_body,
            {
                "job_submission": {
                    "problem_config": {
                        "graph_partitioning": {
                            "graph_file_id": "graph_fid",
                        }
                    },
                    "device_config": {
                        "dirac-1": {
                            "num_samples": 42,
                        }
                    }
                }
            }
        )

    def test_print_job_log(self):
        """Test printing of job log messages."""
        string_io = io.StringIO()
        with redirect_stdout(string_io):
            self.q1.print_job_log(message="fake message")
        out = string_io.getvalue()
        date_now = str(datetime.now().strftime("%Y/%m/%d %H:%M:%S"))[0:10]
        self.assertTrue(out.startswith(f"fake message: {date_now}"))


class TestMultipartUpload(unittest.TestCase):
    """Multipart-file upload test suite."""

    @classmethod
    def setUpClass(cls):
        cls.q1 = QciClient()

    # TODO This client no longer uploads results files, only test download.
    def test_multipart_results_upload_and_download(self):
        """Test uploading/downloading multipart qubo results file."""
        # BIG number of samples, each one num_vars in length
        num_vars = 1700
        num_samp = 300
        is_results_file = True
        samples = np.ones((num_samp, num_vars)).astype(int)
        counts = np.ones((num_samp, 1)).astype(int)
        energies = np.ones((num_samp, 1))

        resdata = {
            "file_name": "test-file.json",
            "file_type": "job_results_sample_qubo",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            "counts": counts.flatten().tolist(),
            "energies": energies.flatten().tolist(),
            "samples": samples.tolist(),
        }

        step_len = compute_results_step_len(resdata["samples"][0])
        expected_parts = ceil(num_samp / step_len)
        self.assertGreater(expected_parts, 1)

        #resp = self.q1.upload_file(file=resdata)
        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #test_res_whole = self.q1.get_file_whole(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(len(test_res_whole["counts"]), counts.shape[0])
        #self.assertEqual(len(test_res_whole["energies"]), energies.shape[0])
        #self.assertEqual(len(test_res_whole["samples"]), samples.shape[0])
        #self.assertEqual(len(test_res_whole["samples"][0]), samples.shape[1])

        #del_resp = self.q1.delete_file(resp["file_id"])
        #assert del_resp["num_deleted"] == 1

    # TODO This client no longer uploads results files, only test download.
    def test_multipart_results_upload_and_download_hamiltonian(self):
        """Test uploading/downloading multipart hamiltonian results file."""
        # Tests float uploads, so we use Hamiltonian job type so the API can handle floats
        num_vars = 20000
        num_samp = 30
        samples = np.ones((num_samp, num_vars)).astype(float)
        counts = np.ones((num_samp, 1)).astype(int)
        energies = np.ones((num_samp, 1))
        is_results_file = True

        resdata = {
            "file_name": "test-file.json",
            "file_type": "job_results_sample_hamiltonian",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            #"solution_type": "continuous",
            "counts": counts.flatten().tolist(),
            "energies": energies.flatten().tolist(),
            "samples": samples.tolist(),
        }

        step_len = compute_results_step_len(resdata["samples"][0])
        expected_parts = ceil(num_samp / step_len)
        self.assertGreater(expected_parts, 1)

        # resp = self.q1.upload_file(file=resdata)
        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #test_res_whole = self.q1.get_file_whole(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(len(test_res_whole["counts"]), len(test_vec_int))
        #self.assertEqual(len(test_res_whole["energies"]), test_vec.shape[0])
        #self.assertEqual(len(test_res_whole["samples"]), num_samples)
        #self.assertEqual(len(test_res_whole["samples"][0]), num_nodes)

        # del_resp = self.q1.delete_file(resp["file_id"])
        # assert del_resp["num_deleted"] == 1

    # TODO This client no longer uploads results files, only test download.
    def test_multipart_results_upload_and_download_huge(self):
        """Test uploading/downloading very large multipart qubo results file."""
        # Now try something too large
        # BIG number of samples, each one num_vars in length
        num_vars = 100000
        num_samp = 300
        samples = np.ones((num_samp, num_vars)).astype(int)
        counts = np.ones((num_samp, 1)).astype(int)
        energies = np.ones((num_samp, 1))
        is_results_file = True

        resdata = {
            "file_name": "test-file.json",
            "file_type": "job_results_sample_qubo",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            "counts": counts.flatten().tolist(),
            "energies": energies.flatten().tolist(),
            "samples": samples.tolist(),
        }

        step_len = compute_results_step_len(samples[0])
        expected_parts = ceil(num_samp / step_len)
        self.assertGreater(expected_parts, 1)

        # resp = self.q1.upload_file(file=resdata)
        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #test_res_whole = self.q1.get_file_whole(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(len(test_res_whole["counts"]), counts.shape[0])
        #self.assertEqual(len(test_res_whole["energies"]), energies.shape[0])
        #self.assertEqual(len(test_res_whole["samples"]), samples.shape[0])
        #self.assertEqual(len(test_res_whole["samples"][0]), samples.shape[1])


    # TODO This client no longer uploads results files, only test download.
    def test_multipart_graph_partitioning_results_upload_and_download(self):
        """Test uploading/downloading multipart graph-partitioning results file."""
        num_samples = 10
        num_nodes = 10000
        test_vec = np.arange(num_samples)
        test_vec_int = np.arange(num_samples).astype(int)
        is_results_file = True

        samples = [
            [
                {"id": int(np.random.randint(0, 10)), "class": np.random.randint(0, 3)}
                for _ in range(num_nodes)
            ]  # pretend we have a graph with 100 nodes
            for _ in range(num_samples)  # and we asked for 1000 samples
        ]

        gp_results = {
            "file_name": "test-file.json",
            "file_type": "job_results_graph_partitioning",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            "balance": test_vec.tolist(),
            "counts": test_vec.tolist(),
            "cut_size": test_vec_int.tolist(),
            "energies": test_vec.tolist(),
            "is_feasible": [True] * num_samples,
            "samples": samples,
        }

        step_len = compute_results_step_len(samples[0])
        expected_parts = ceil(num_samples / step_len)
        self.assertGreater(expected_parts, 1)

        # resp = self.q1.upload_file(file=gp_results)

        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #del_resp = self.q1.delete_file(resp["file_id"])
        #assert del_resp["num_deleted"] == 1


@pytest.mark.offline
class TestJobsApiWithRequestMocks(unittest.TestCase):
    """Jobs-API-related test suite that can be run without backend."""

    def setUp(self):
        with pytest.MonkeyPatch().context() as mp:
            mp.setenv('QCI_TOKEN', 'test_api_token')
            mp.setenv('QCI_API_URL', 'test_url')
            self.qci_client = QciClient(set_bearer_token_on_init=False)
        self.job_id = "63b717a22da68618ec444eac"
        self.job_type = "sample-hamiltonian"
        self.file_id = "73b717a22da68618ec444eab"

        # Bad GET jobs response.
        self.get_response_bad = requests.Response()
        self.get_response_bad.status_code = 404

    def test_jobs_url(self) -> None:
        """Test getting jobs URL."""
        self.assertEqual(self.qci_client.jobs_url, "test_url/optimization/v1/jobs")

    def test_get_job_id_url(self) -> None:
        """Test getting jobs URL for given job ID."""
        self.assertEqual(
            self.qci_client.get_job_id_url(self.job_id), f"test_url/optimization/v1/jobs/{self.job_id}"
        )

    def test_get_job_status_url(self) -> None:
        """Test getting jobs-status URL for given job ID."""
        self.assertEqual(
            self.qci_client.get_job_status_url(self.job_id),
            f"test_url/optimization/v1/jobs/{self.job_id}/status",
        )

    def test_files_url(self) -> None:
        """Test getting files URL."""
        self.assertEqual(self.qci_client.files_url, "test_url/optimization/v1/files")

    def test_get_file_id_url(self) -> None:
        """Test getting files URL for given file ID."""
        self.assertEqual(
            self.qci_client.get_file_id_url(self.file_id),
            f"test_url/optimization/v1/files/{self.file_id}",
        )

    def test_get_job_type_from_job_id(self) -> None:
        """Test getting job type from job ID."""
        # GET short jobs response.
        get_jobs_response_short = requests.Response()
        get_jobs_response_short.status_code = 200
        get_jobs_response_short._content = (  # pylint: disable=protected-access
            b"""{
    "job_id": "%b",
    "status": "COMPLETED",
    "type": "sample_hamiltonian",
    "organization_id": "6edf5db3def87d53b6bf375b",
    "username": "test_user"
}"""
            % self.job_id.encode()
        )

        self.qci_client.session.request = unittest.mock.MagicMock(
            return_value=get_jobs_response_short
        )
        self.assertEqual(
            self.qci_client.get_job_type_from_job_id(self.job_id), self.job_type
        )

        self.qci_client.session.request = unittest.mock.MagicMock(
            return_value=self.get_response_bad
        )
        with self.assertRaises(AssertionError):
            self.qci_client.get_job_type_from_job_id(self.job_id)



"""
QciClient
Utility class for user interactions with QCI API
"""

import concurrent.futures
from dataclasses import dataclass
from datetime import datetime
import gzip
from io import BytesIO
import json
from posixpath import join
import time
from typing import ClassVar, Optional

from requests.adapters import HTTPAdapter, Retry
from requests_futures.sessions import FuturesSession

from qci_client import utilities
from qci_client.base import BaseApi, BACKOFF_FACTOR, RETRY_TOTAL, STATUS_FORCELIST
from qci_client.data_converter import data_to_json

TIMEOUT_DEFAULT: Optional[float] = 2 * 60.0  # seconds, or None for infinite.
COMPRESS_DEFAULT = False
MAX_WORKERS = 8

class JobStatus:  # pylint: disable=too-few-public-methods
    """Allowed jobs statuses."""

    QUEUED = "QUEUED"
    SUBMITTED = "SUBMITTED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    ERRORED = "ERRORED"
    CANCELLED = "CANCELLED"

FINAL_STATUSES = frozenset([JobStatus.COMPLETED, JobStatus.ERRORED, JobStatus.CANCELLED])

@dataclass
class QciClient(BaseApi):  # pylint: disable=too-many-public-methods
    """
    Provides requests for QCi's public API for running optimization problems on Dirac
    devices, including file uploads/downloads and submitting/retrieving entire jobs.

    Accepts same parameters as :class:`qci_client.base.BaseApi` as well as:

    :param files: url path fragment to specify files API endpoint
    :param jobs: url path fragment to specify jobs API endpoint
    :param max_workers: number of threads for concurrent file download calls
    :param compress: compress file metadata and parts before uploading
    """
    files: str = "optimization/v1/files"
    jobs: str = "optimization/v1/jobs"
    max_workers: int = 8
    compress: bool = COMPRESS_DEFAULT

    _supported_job_types: ClassVar[frozenset] = frozenset(
        [
            "sample-qubo",
            "graph-partitioning",
            "sample-constraint",
            "sample-hamiltonian",
            "sample-hamiltonian-ising",
        ]
    )

    @property
    def jobs_url(self):
        """Get jobs URL."""
        return join(self.url, self.jobs)

    def get_job_id_url(self, job_id: str) -> str:
        """Get job URL with job ID."""
        return join(self.jobs_url, job_id)

    def get_job_status_url(self, job_id: str) -> str:
        """Get job status using job ID."""
        return join(self.get_job_id_url(job_id), "status")

    def get_job_allocations_url(self) -> str:
        """Get job allocations"""
        return join(self.jobs_url, "allocations")

    @property
    def files_url(self):
        """Get files URL."""
        return join(self.url, self.files)

    def get_file_id_url(self, file_id: str) -> str:
        """Get file URL with file ID."""
        return join(self.files_url, file_id)

    def get_file_contents_url(self, file_id: str, part_num: int) -> str:
        """Get file contents URL with file ID and file part number."""
        return join(self.get_file_id_url(file_id), "contents", str(part_num))

    @property
    def headers_without_authorization(self) -> dict:
        """HTTP headers without bearer token."""
        headers = {
            "Content-Type": "application/json",
            # Simple, sessionless requests, so close connection proactively.
            "Connection": "close",
        }

        if self.timeout is not None:
            # Tell server when client will stop waiting for response.
            headers["X-Request-Timeout-Nano"] = str(int(10**9 * self.timeout))

        return headers

    @BaseApi.refresh_token
    def upload_file(  # pylint: disable=too-many-branches, too-many-locals, too-many-statements
        self,
        file: dict,
    ) -> dict:
        """
        Upload file (metadata and then parts concurrently). Returns dict with file ID.
        """
        # Use session with maintained connection and multipart concurrency for
        # efficiency.
        file = data_to_json(file=file)

        with FuturesSession(max_workers=self.max_workers) as session:
            session.mount(
                "https://",
                HTTPAdapter(
                    max_retries=Retry(
                        total=RETRY_TOTAL,
                        backoff_factor=BACKOFF_FACTOR,
                        status_forcelist=STATUS_FORCELIST,
                    )
                ),
            )

            post_response_future = session.post(
                self.files_url,
                headers=self.headers_without_connection_close,
                timeout=self.timeout,
                json=utilities.get_post_request_body(file=file),
            )

            for response_future in concurrent.futures.as_completed(
                [post_response_future], self.timeout
            ):
                response = response_future.result()
                response.raise_for_status()

            file_id = response.json()["file_id"]
            file_part_generator = utilities.file_part_generator(
                file=file, compress=self.compress
            )
            patch_response_futures = []

            if self.compress:
                for part_body, part_number in file_part_generator:
                    patch_response_futures.append(
                        session.patch(
                            join(self.files_url, f"{file_id}/contents/{part_number}"),
                            headers=self.headers_without_connection_close,
                            timeout=self.timeout,
                            data=utilities.zip_payload(
                                payload=utilities.get_patch_request_body(file=part_body)
                            ),
                        )
                    )
            else:
                for part_body, part_number in file_part_generator:
                    patch_response_futures.append(
                        session.patch(
                            join(self.files_url, f"{file_id}/contents/{part_number}"),
                            headers=self.headers_without_connection_close,
                            timeout=self.timeout,
                            json=utilities.get_patch_request_body(file=part_body),
                        )
                    )

            # Due to timeout in underlying PATCH, this should not hang despite no
            # timeout.
            for response_future in concurrent.futures.as_completed(
                patch_response_futures
            ):
                response = response_future.result()
                response.raise_for_status()

        return {"file_id": file_id}

    @BaseApi.refresh_token
    def download_file(self, *, file_id: str) -> dict:
        """Download file (metadata and then parts concurrently)."""
        # Use session with maintained connection and multipart concurrency for
        # efficiency.
        with FuturesSession(max_workers=self.max_workers) as session:
            session.mount(
                "https://",
                HTTPAdapter(
                    max_retries=Retry(
                        total=RETRY_TOTAL,
                        backoff_factor=BACKOFF_FACTOR,
                        status_forcelist=STATUS_FORCELIST,
                    )
                ),
            )

            get_response_future = session.get(
                #urljoin(self.files_url, file_id),
                join(self.files_url, file_id),
                headers=self.headers_without_connection_close,
                timeout=self.timeout,
            )

            for response_future in concurrent.futures.as_completed(
                [get_response_future], self.timeout
            ):
                response = response_future.result()
                response.raise_for_status()

            # File metadata is base for returned fully assembled file.
            file = {**response.json()}

            # Remove metadata fields that are not well-defined for fully assembled file.
            file.pop("last_accessed_rfc3339")
            file.pop("upload_date_rfc3339")

            get_response_futures = [
                session.get(
                    #urljoin(self.files_url, f"{file_id}/contents/{part_number}"),
                    join(self.files_url, f"{file_id}/contents/{part_number}"),
                    headers=self.headers_without_connection_close,
                    timeout=self.timeout,
                )
                for part_number in range(1, file["num_parts"] + 1)
            ]

            # Due to timeout in underlying GET, this should not hang despite no timeout.
            for response_future in concurrent.futures.as_completed(
                get_response_futures
            ):
                response = response_future.result()
                response.raise_for_status()

            # Unpack in order.
            for response_future in get_response_futures:
                file_part = response_future.result().json()
                # Append to all array fields.
                for file_type, file_type_config in file_part["file_config"].items():
                    if file_type not in file["file_config"]:
                        file["file_config"][file_type] = {}

                    for key, value in file_type_config.items():
                        if key not in file["file_config"][file_type]:
                            file["file_config"][file_type][key] = []

                        file["file_config"][file_type][key] += value

        return file

    @BaseApi.refresh_token
    def submit_job(self, job_body: dict, job_type: str) -> dict:
        """
        Submit a job via a request to QCI public API.

        Args:
            job_body: formatted json body that includes all parameters for the job
            job_type: one of the _supported_job_types

        Returns:
            Response from POST call to API
        """
        self.validate_job_type(job_type=job_type)
        response = self.session.request(
            "POST",
            self.jobs_url,
            json=job_body,
            headers=self.headers,
            timeout=self.timeout,
        )
        self._check_response_error(response=response)
        return response.json()

    @BaseApi.refresh_token
    def get_job_status(self, job_id: str) -> dict:
        """
        Get the status of a job by its ID.

        Args:
            job_id: ID of job

        Returns:
            Response from GET call to API
        """
        response = self.session.request(
            "GET",
            self.get_job_status_url(job_id),
            headers=self.headers,
            timeout=self.timeout,
        )

        self._check_response_error(response=response)
        return response.json()

    @BaseApi.refresh_token
    def get_job_response(self, job_id: str, job_type: str) -> dict:
        """
        Get a response for a job by id and type, which may/may not be finished.

        :param job_id: ID of job
        :param job_type: type of job, one of []

        :return dict: loaded json file
        """
        self.validate_job_type(job_type=job_type)
        response = self.session.request(
            "GET",
            self.get_job_id_url(job_id),
            headers=self.headers,
            timeout=self.timeout,
        )

        self._check_response_error(response=response)
        return response.json()

    def validate_job_type(self, job_type: str) -> None:
        """
        Checks if a provided job type is a supported job type.

        Args:
            job_type: a job type to validate

        Returns:
            None

        Raises AssertionError if job_type is not one of the _supported_job_types.
        """
        if job_type not in self._supported_job_types:
            raise AssertionError(
                f"Provided job_type '{job_type}' is not one of "
                f"{self._supported_job_types}"
            )

    def build_job_body(  # pylint: disable=too-many-arguments
        self,
        job_type: str,
        job_params: dict,
        qubo_file_id: Optional[str] = None,
        graph_file_id: Optional[str] = None,
        hamiltonian_file_id: Optional[str] = None,
        objective_file_id: Optional[str] = None,
        constraints_file_id: Optional[str] = None,
        polynomial_file_id: Optional[str] = None,
        job_name: Optional[str] = None,
        job_tags: Optional[list] = None,
    ) -> dict:
        """
        Constructs body for job submission requests

        Args:
            job_type: one of _supported_job_types
            job_params: dict of params to be passed to job submission in "params" key
            qubo_file_id: file id from files API for uploaded qubo
            graph_file_id: file id from files API for uploaded graph
            hamiltonian_file_id: file id from files API for uploaded hamiltonian
            objective_file_id: file id from files API for uploaded objective
            constraints_file_id: file id from files API for uploaded constraints
            job_name: user specified name for job submission
            job_tags: user specified labels for classifying and filtering user jobs after submission

        Returns:
            None
        """
        # TODO: Need to add validation for job parameters
        self.validate_job_type(job_type=job_type)

        problem_config = {}
        device_config = {}

        if "sampler_type" not in job_params:
            raise ValueError(
                "Must define sampler_type in job_params (dirac-1, dirac-2, or dirac-3)."
            )

        device_name = job_params['sampler_type']

        # TODO: remove this in the future for now map dirac- to eqc but warn of deprecation
        if "eqc" in device_name:
            print("WARNING: " + device_name + " will be a deprecated sampler type dirac-(1-3) will be the supported sampler types in the future")
            device_name = device_name.replace("eqc", "dirac-")

        # Optional nsamples.
        num_samples = job_params.get("nsamples")

        if num_samples is None:
            # Fallback to checking deprecated fields.
            if "n_samples" in job_params:
                print("WARNING: the key n_samples will be a deprecated parameter in the future, nsamples will be the supported parameter.")
                num_samples = job_params["n_samples"]
            elif "num_samples" in job_params:
                print("WARNING: the key num_samples will be a deprecated parameter in the future, nsamples will be the supported parameter.")
                num_samples = job_params["num_samples"]
            elif "num_solutions" in job_params:
                print("WARNING: the key num_solutions will be a deprecated parameter in the future, nsamples will be the supported parameter.")
                num_samples = job_params["num_solutions"]

        if num_samples is not None:
            # Optional parameter.
            device_config["num_samples"] = num_samples

        if job_type == 'sample-qubo':
            if device_name in ('dirac-2', 'dirac-3'):
                raise ValueError(
                    f"{job_type} not supported on dirac-2 and dirac-3. Consider using "
                    "job_type sample-hamiltonian for dirac-2 and dirac-3."
                )

            problem_name = 'quadratic_unconstrained_binary_optimization'

            if not qubo_file_id:
                raise AssertionError(
                    "qubo_file_id must be specified for job_type='sample-qubo'"
                )

            problem_config["qubo_file_id"] = qubo_file_id
        elif job_type == 'sample-hamiltonian':
            if device_name == 'dirac-2':
                if job_params.get("solution_precision") != 1:
                    problem_name = 'normalized_qudit_hamiltonian_optimization_continuous'
                else:
                    problem_name = 'normalized_qudit_hamiltonian_optimization_integer'
            elif device_name == 'dirac-3':
                problem_name = 'normalized_qudit_hamiltonian_optimization'

                if "relaxation_schedule" in job_params:
                    # Optional parameter.
                    device_config["relaxation_schedule"] = job_params["relaxation_schedule"]

                if "solution_precision" in job_params:
                    # Optional parameter.
                    device_config["solution_precision"] = job_params["solution_precision"]

                if "sum_constraint" in job_params:
                    # Optional parameter.
                    device_config["sum_constraint"] = job_params["sum_constraint"]
            else:
                raise ValueError(f"{job_type} not supported on {device_name}.")

            if (not hamiltonian_file_id and not polynomial_file_id) or \
                (hamiltonian_file_id and polynomial_file_id):
                raise AssertionError(
                    "exactly one of hamiltonian_file_id or polynomial_file_id must be "
                    "specified for job_type='sample-hamiltonian'"
                )

            if hamiltonian_file_id:
                problem_config["hamiltonian_file_id"] = hamiltonian_file_id
            else:
                problem_config["polynomial_file_id"] = polynomial_file_id
        elif job_type == 'sample-hamiltonian-ising':
            if device_name in ('dirac-2', 'dirac-3'):
                raise ValueError(
                    f"{job_type} not supported on dirac-2 and dirac-3. Consider using "
                    "job_type sample-hamiltonian for dirac-2 and dirac-3."
                )

            problem_name = 'ising_hamiltonian_optimization'
            
            if (not hamiltonian_file_id and not polynomial_file_id) or \
                (hamiltonian_file_id and polynomial_file_id):
                raise AssertionError(
                    "exactly one of hamiltonian_file_id or polynomial_file_id must be "
                    "specified for job_type='sample-hamiltonian-ising'"
                )

            if hamiltonian_file_id:
                problem_config["hamiltonian_file_id"] = hamiltonian_file_id
            else:
                problem_config["polynomial_file_id"] = polynomial_file_id
        elif job_type == "sample-constraint":
            if device_name in ('dirac-2', 'dirac-3'):
                raise ValueError(
                    f"{job_type} not supported on dirac-2 and dirac-3. Consider using "
                    "job_type sample-hamiltonian for dirac-2 and dirac-3."
                )

            problem_name = 'quadratic_linearly_constrained_binary_optimization'

            if not constraints_file_id:
                raise AssertionError(
                    "At least constraints_file_id must be specified for "
                    "job_type='sample-constraint'"
                )

            problem_config["constraints_file_id"] = constraints_file_id
            problem_config["objective_file_id"] = objective_file_id  # Optional.

            if "alpha" in job_params:
                # Optional parameter.
                problem_config["alpha"] = job_params["alpha"]

            if "atol" in job_params:
                # Optional parameter.
                problem_config["atol"] = job_params["atol"]
        elif job_type == "graph-partitioning":
            if device_name in ('dirac-2', 'dirac-3'):
                raise ValueError(
                    f"{job_type} not supported on dirac-2 and dirac-3. Consider using "
                    "job_type sample-hamiltonian for dirac-2 and dirac-3."
                )

            problem_name = 'graph_partitioning'

            if not graph_file_id:
                raise AssertionError(
                    "graph_file_id must be specified for the given "
                    "job_type='graph-partitioning'"
                )

            problem_config["graph_file_id"] = graph_file_id

            if "num_paritions" in job_params:
                # Optional parameter.
                problem_config["num_paritions"] = job_params["num_paritions"]

            if "alpha" in job_params:
                # Optional parameter.
                problem_config["alpha"] = job_params["alpha"]

            if "gamma" in job_params:
                # Optional parameter when num_paritions > 2.
                problem_config["gamma"] = job_params["gamma"]
        else:
            raise ValueError(f"Unsupported job_type '{job_type}'.")

        job_submission = {
            "problem_config": {problem_name: problem_config},
            "device_config": {device_name: device_config},
        }

        if job_name is not None:
            # Optional field.
            job_submission["job_name"] = job_name

        if job_tags is not None:
            # Optional field.
            job_submission["job_tags"] = job_tags

        return {"job_submission": job_submission}

    def print_job_log(self, message: str) -> None:
        """
        Formats a messages for updating user with a time stamp appended
        :param message: a string to be passed in print statement
        """
        print(f"{message}: {datetime.now().strftime('%Y/%m/%d %H:%M:%S')}")

    def process_job(self, job_type: str, job_body: dict, wait: bool = True) -> dict:
        """
        :param job_type: the type of job being processed must be one of _supported_job_types
        :param job_body: formatted json dict for body of job submission request
        :param wait: bool indicating whether or not user wants to wait for job to complete

        :return:
            if wait is True, then dict with job_info response and results file
                (results is None if ERRORED or CANCELLED)
            if wait is False, then response dict from submitted job, which includes job
                ID for subsequent retrieval
        :note: what else do we want to return with the results? response_id, obviously job_id
        """
        self.validate_job_type(job_type=job_type)

        allocations_response = self.session.request(
            "GET",
            self.get_job_allocations_url(),
            headers=self.headers,
            timeout=self.timeout,
        ).json()
        dirac_allocation = allocations_response['allocations']['dirac']['seconds']
    
        if allocations_response['allocations']['dirac']['metered']:
            print(f"Dirac allocation balance = {dirac_allocation} s")
        else:
            print(f"Dirac allocation balance = {dirac_allocation} s (unmetered)")
    
        submit_response = self.submit_job(job_body=job_body, job_type=job_type)
        job_id = submit_response["job_id"]
        self.print_job_log(message=f"Job submitted job_id='{job_id}'-")

        if wait:
            curr_status = ""
            while curr_status not in FINAL_STATUSES:
                time.sleep(1)
                iter_status = self.get_job_status(job_id=job_id)["status"]

                if iter_status != curr_status:
                    self.print_job_log(message=iter_status)
                    curr_status = iter_status

            job_response = self.get_job_response(job_id=job_id, job_type=job_type)
            job_response['details'] = {'status': curr_status}

            if curr_status in [JobStatus.CANCELLED, JobStatus.ERRORED]:
                results = None
            else:
                results = self.download_file(file_id=job_response["job_result"]["file_id"])

            allocations_response = self.session.request(
                "GET",
                self.get_job_allocations_url(),
                headers=self.headers,
                timeout=self.timeout,
            ).json()
            dirac_allocation = allocations_response['allocations']['dirac']['seconds']
        
            if allocations_response['allocations']['dirac']['metered']:
                print(f"Dirac allocation balance = {dirac_allocation} s")
            else:
                print(f"Dirac allocation balance = {dirac_allocation} s (unmetered)")

            return {"job_info": job_response, "results": results}

        return submit_response

    @BaseApi.refresh_token
    def list_files(self, username: Optional[str] = None) -> dict:
        """
        :param username: Optional str - username (to search for files owned by the named user)
            mostly useful when run by users with administrator privileges (such as QCI users) who can see all files.
            When called by an administrator, the username parameter is used to restrict the list files returned
            to be only the files owned by the user specified in the username parameter.
            When run by non-privileged users, this parameter is truly optional because non-privileged users
            will only ever see lists of files that they created.

        :return: dict containing list of files
        """
        if username:
            querystring = {"regname": "username", "regvalue": username}

            response = self.session.request(
                "GET",
                self.files_url,
                headers=self.headers,
                params=querystring,
                timeout=self.timeout,
            )
        else:
            response = self.session.request(
                "GET",
                self.files_url,
                headers=self.headers,
                timeout=self.timeout,
            )

        self._check_response_error(response=response)
        return response.json()

    @BaseApi.refresh_token
    def delete_file(self, file_id: str) -> dict:
        """
        :param file_id: str - file_id of file to be deleted

        :return: dict containing information about file deleted (or error)
        """

        if self.debug:
            print(f"Deleting file with ID {file_id}...")

        start_time_s = time.perf_counter()

        response = self.session.request(
            "DELETE",
            self.get_file_id_url(file_id),
            headers=self.headers,
            timeout=self.timeout,
        )

        stop_time_s = time.perf_counter()

        if self.debug:
            print(f"Deleting file with ID {file_id}...done.")
            print(f"\tElapsed time: {stop_time_s - start_time_s} s.")

        self._check_response_error(response=response)

        return response.json()

    @BaseApi.refresh_token
    def zip_payload(self, payload: str) -> bytes:
        """
        Zip contents of json file

        Args:
            payload: str - json contents of file to be zipped

        Returns:
            zipped request_body
        """
        out = BytesIO()
        with gzip.GzipFile(fileobj=out, mode="w", compresslevel=6) as file:
            file.write(json.dumps(payload).encode("utf-8"))
        request_body = out.getvalue()
        out.close()
        return request_body

    def get_job_type_from_job_id(self, job_id: str) -> str:
        """
        Get job type from job ID.

        Args:
            job_id: ID of the job

        Returns:
            Type of the job
        """
        response_job_metadata_short = self.session.request(
            "GET",
            self.get_job_id_url(job_id),
            headers=self.headers,
            timeout=self.timeout,
        )
        self._check_response_error(response=response_job_metadata_short)

        return response_job_metadata_short.json()["type"].replace("_", "-")



"""Functions for data conversion."""

from math import floor
import sys
import time
from typing import Union

import networkx as nx
import numpy as np
import scipy.sparse as sp

from qci_client import enum

# We want to limit the memory size of each uploaded chunk to be safely below the max of 15 * MebiByte (~15MB).
# See https://git.qci-dev.com/qci-dev/qphoton-files-api/-/blob/main/service/files.go#L32.
MEMORY_MAX: int = 8 * 1000000  # 8MB


def get_size(obj, seen=None) -> int:
    """
    Recursively finds size of objects

    :param obj: data object to recursively compute size of
    :param seen: takes a set and is used in the recursive step only to record whether an object has been counted yet.

    :return int:
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum(get_size(v, seen) for v in obj.values())
        size += sum(get_size(k, seen) for k in obj.keys())
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum(get_size(i, seen) for i in obj)
    return size


def _get_soln_size(soln):
    # Check whether first entry is a graph node/class assignment, eg., {'id': 4, 'class': 2}
    if isinstance(soln[0], dict):
        return get_size(soln)

    return sys.getsizeof(soln[0]) * len(soln)


def compute_results_step_len(data: Union[np.ndarray, list]) -> int:
    """
    Compute the step length for "chunking" the providd data.

    Args:
        data: An numpy array or list of data

    Returns:
        The step length for "chunking" the data
    """
    # total mem size of soln vector
    soln_mem = _get_soln_size(data)
    # num_vars * step_len < 30k => step_len < 30k/num_vars
    chunk_ratio = MEMORY_MAX / soln_mem
    step_len = floor(chunk_ratio) if chunk_ratio >= 1 else 1
    return step_len

def data_to_json(file: dict, debug: bool = False) -> dict:
    """
    Converts data in file input into JSON-serializable dictionary that can be passed to Qatalyst REST API

    Args:
        file: file dictionary whose data of type numpy.ndarray, scipy.sparse.spmatrix, or networkx.Graph is to be converted
        debug: Optional, if set to True, enables debug output (default = False for no debug output)

    Returns:
        file dictionary with JSON-serializable data
    """
    start_time_s = time.perf_counter()

    supported_file_types = [type.value for type in enum.JOB_INPUTS_FILE_TYPES]
    supported_file_types.sort()
    supported_file_types = tuple(supported_file_types)
    matrix_file_types = [type.value for type in enum.JOB_INPUTS_MATRIX_FILE_TYPES]
    matrix_file_types.sort()
    matrix_file_types = tuple(matrix_file_types)

    file_type = enum.get_file_type(file=file).value

    if file_type not in supported_file_types:
        raise AssertionError(
            f"data conversion not supported for file type '{file_type}', supported "
            f"types are {supported_file_types}"
        )

    data = file['file_config'][file_type]['data']

    if file_type == "graph":
        if not isinstance(data, nx.Graph):
            raise AssertionError("file_type 'graph' data must be a networkx.Graph")

        file_config = {
            **nx.node_link_data(data),
            "num_edges": data.number_of_edges(),
            "num_nodes": data.number_of_nodes(),
        }
    elif file_type in matrix_file_types:
        if isinstance(data, nx.Graph):
            raise AssertionError(
                f"file_type '{file_type}' data cannot be a networkx.Graph"
            )

        data_ls = []

        if sp.isspmatrix_dok(data):
            for idx, val in zip(data.keys(), data.values()):
                # dok type has trouble subsequently serializing to json without type
                # casts. For example, uint16 and float32 cause problems.
                data_ls.append({"i": int(idx[0]), "j": int(idx[1]), "val": float(val)})
        elif sp.isspmatrix(data) or isinstance(data, np.ndarray):
            data = sp.coo_matrix(data)

            for i, j, val in zip(
                data.row.tolist(), data.col.tolist(), data.data.tolist()
            ):
                data_ls.append({"i": i, "j": j, "val": val})
        else:
            raise ValueError(
                f"file_type '{file_type}' only supports types numpy.ndarray and "
                f"scipy.sparse.spmatrix, got {type(data)}"
            )

        file_config = {"data": data_ls}
        rows, cols = data.get_shape()

        if file_type == "constraints":
            # Constraints matrix is [A | -b]
            file_config.update({"num_constraints": rows, "num_variables": cols-1})
        else:
            # This works for hamiltonians, qubos, and objectives.
            file_config["num_variables"] = rows
    else:
        # Polynomial file types do not require translation.
        file_config = file["file_config"][file_type]

    if debug:
        print(f"Time to convert data to json: {time.perf_counter()-start_time_s} s.")

    return {
        "file_name": file.get("file_name", f"{file_type}.json"),
        "file_config": {file_type: file_config}
    }
